/*
 * Quoridor server and related libraries
 * Copyright (C) 2009-2010 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://www.gnu.org/licenses/.
 */
package cz.xelfi.quoridor.freemarkerdor;

import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.UniformInterfaceException;
import com.sun.jersey.api.client.WebResource;
import com.sun.jersey.api.container.httpserver.HttpServerFactory;
import com.sun.jersey.api.core.PackagesResourceConfig;
import com.sun.jersey.api.core.ResourceConfig;
import com.sun.jersey.api.view.Viewable;
import com.sun.net.httpserver.HttpServer;
import cz.xelfi.quoridor.Board;
import cz.xelfi.quoridor.IllegalPositionException;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.URI;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.concurrent.Callable;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.FormParam;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.CacheControl;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Cookie;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.NewCookie;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import org.openid4java.consumer.ConsumerManager;
import org.openid4java.consumer.VerificationResult;
import org.openid4java.discovery.DiscoveryInformation;
import org.openid4java.message.AuthRequest;
import org.openid4java.message.MessageExtension;
import org.openid4java.message.Parameter;
import org.openid4java.message.ParameterList;
import org.openid4java.message.ax.AxMessage;
import org.openid4java.message.ax.FetchRequest;
import org.openid4java.message.ax.FetchResponse;
import org.openide.util.Exceptions;
import org.w3c.dom.Document;

/**
 *
 * @author Jaroslav Tulach <jtulach@netbeans.org>
 */
@Path("/")
public final class UI {
    private static final String version;
    static {
        Properties p = new Properties();
        try {
            InputStream is = FreemarkerProcessor.class.getResourceAsStream("/META-INF/maven/cz.xelfi.quoridor/freemarkerdor/pom.properties"); // NOI18N
            if (is != null) {
                p.load(is);
            }
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        version = p.getProperty("version", "unknown"); // NOI18N
    }
    private static final Logger LOG = Logger.getLogger(UI.class.getName());
    private static WebResource base;
    private static WebResource stat;
    private static WebResource web;
    private static Requests requests;
    private static ConsumerManager manager;
    private static Map<String,DiscoveryInformation> ids = new HashMap<String, DiscoveryInformation>();

    @Context
    private HttpHeaders headers;
    private UserInfo user;
    private String uuid;

    public UI() {
    }

    private String login() {
        Cookie cookie = headers.getCookies().get("login");
        if (cookie != null) {
            return cookie.getValue();
        }
        return null;
    }

    private Viewable checkLogin() {
        String id = login();
        if (id != null) {
            UserInfo us;
            try {
                us = base.path("users").queryParam("loginID", id).
                    accept(MediaType.TEXT_XML).get(UserInfo.class);
            } catch (Exception ex) {
                ex.printStackTrace();
                us = null;
            }
            if (us != null && us.getId().length() > 0) {
                user = us;
                uuid = id;
                return null;
            }
        }
        return viewable("login.fmt", null);
    }

    @GET
    @Path("openid")
    @Produces(MediaType.TEXT_HTML)
    public Viewable openidResponse(
        @QueryParam("openid.assoc_handle") String handle,
        @QueryParam("openid.claimed_id") String claimedID,
        @QueryParam("openid.identity") String identity,
        @QueryParam("openid.mode") String mode /*id_res */,
        @QueryParam("openid.ns") String ns, /* http%3A%2F%2Fspecs.openid.net%2Fauth%2F2.0 */
        @QueryParam("openid.op_endpoint") String endpoint,
        @QueryParam("openid.response_nonce") String nonce,
        @QueryParam("openid.return_to") String returnTo,
        @QueryParam("openid.sig") String sig,
        @QueryParam("openid.signed") String signed
    ) throws Exception {
        if (signed == null) {
            return viewable("openid.fmt", null);
        }
        
        if (manager == null) {
            manager = new ConsumerManager(); 
        }
        ParameterList pl = new ParameterList();
        pl.set(new Parameter("openid.assoc_handle", handle));
        pl.set(new Parameter("openid.claimed_id", claimedID));
        pl.set(new Parameter("openid.identity", identity));
        pl.set(new Parameter("openid.mode", mode));
        pl.set(new Parameter("openid.ns", ns));
        pl.set(new Parameter("openid.op_endpoint", endpoint));
        pl.set(new Parameter("openid.response_nonce", nonce));
        pl.set(new Parameter("openid.return_to", returnTo));
        pl.set(new Parameter("openid.sig", sig));
        pl.set(new Parameter("openid.signed", signed));
        
        DiscoveryInformation info = ids.get(claimedID);
        VerificationResult res = manager.verify(returnTo, pl, info);
        String userId = res != null && res.getVerifiedId() != null ? 
            res.getVerifiedId().getIdentifier() : null;
        
        if (res.getAuthResponse().hasExtension(AxMessage.OPENID_NS_AX)) {
            MessageExtension ext = res.getAuthResponse().getExtension(AxMessage.OPENID_NS_AX);

            if (ext instanceof FetchResponse) {
                FetchResponse fetchResp = (FetchResponse) ext;

                String firstName = fetchResp.getAttributeValue("FirstName");
                String lastName = fetchResp.getAttributeValue("LastName");
                String email = fetchResp.getAttributeValue("Email");
                
                LOG.log(Level.INFO, "First name: {0}", firstName);
                LOG.log(Level.INFO, "Last name: {0}", lastName);
                LOG.log(Level.INFO, "Email: {0}", email);
                
                
            } 
        }        
        
        return viewable("openid.fmt", null, "id", userId);
    }
    @POST
    @Path("openid")
    @Produces(MediaType.TEXT_HTML)
    public Viewable openid(@FormParam("openid_identifier") String openid) throws Exception {
        if (manager == null) {
            manager = new ConsumerManager(); 
        }
        List l = manager.discover(openid);
        DiscoveryInformation info = manager.associate(l);
        AuthRequest auth = manager.authenticate(info, web.path("openid").getURI().toString());
        FetchRequest fetch = FetchRequest.createFetchRequest();
        fetch.addAttribute("FirstName", "http://schema.openid.net/namePerson/first", true);
        fetch.addAttribute("LastName", "http://schema.openid.net/namePerson/last", true);
        fetch.addAttribute("Email", "http://schema.openid.net/contact/email", true);
        auth.addExtension(fetch);        
        ids.put(info.getClaimedIdentifier().toString(), info);
        
        return viewable("openid.fmt", null, "url", auth.getDestinationUrl(true));
    }
    
    @POST
    @Path("login")
    @Produces(MediaType.TEXT_HTML)
    public Response login(
        @FormParam("name") String name, @FormParam("password") String password
    ) throws Exception {
        uuid = base.path("login").queryParam("name", name).queryParam("password", password).
            accept(MediaType.TEXT_PLAIN).put(String.class);
        if (uuid != null) {
            user = new UserInfo(name);
            NewCookie nc = new NewCookie("login", uuid, null, null, null, 3600 * 24 * 7, false);
            return Response.ok().cookie(nc).entity(viewable("login.fmt", null)).build();
        } else {
            Viewable v = viewable("login.fmt", null, "message", "Invalid name or password: " + name);
            return Response.status(1).entity(v).build();
        }
    }

    @GET
    @Produces(MediaType.TEXT_HTML)
    public Response welcome(@QueryParam("maxItems") @DefaultValue("10") int maxItems) {
        Viewable v = checkLogin();
        ResponseBuilder resp = Response.ok();
        if (v == null) {
            v = welcomeImpl("maxItems", maxItems);
        }
        CacheControl cc = new CacheControl();
        cc.setNoCache(true);
        resp.cacheControl(cc);
        return resp.entity(v).build();
    }

    @GET
    @Path("games/{id}.png")
    @Produces("image/png")
    public Response getBoardImage(
        @PathParam("id") String id,
        @QueryParam("fieldSize") @DefaultValue("50") int fieldSize,
        @QueryParam("move") @DefaultValue("-1") int move
    ) throws IllegalPositionException {
        WebResource wr = base.path("games").path(id);
        if (move != -1) {
            wr = wr.queryParam("move", "" + move);
        }
        String txt = wr.accept(MediaType.TEXT_PLAIN).get(String.class);
        Board b = Board.valueOf(txt);
//        Board b = new Board(txt);
        ResponseBuilder resp = Response.ok();
        CacheControl cc = new CacheControl();
        cc.setNoCache(true);
        resp.cacheControl(cc);
        return resp.entity(BoardImage.draw(b, fieldSize)).build();
    }


    private Response board(String id) {
        return board(id, "", null);
    }
    @GET
    @Path("games/{id}/")
    @Produces(MediaType.TEXT_HTML)
    public Response board(
        @PathParam("id") String id,
        @QueryParam("format") @DefaultValue("") String format,
        @QueryParam("move") @DefaultValue("-1") String move
    ) {
        return board(id, null, format, move);
    }
    private Response board(@PathParam("id") String id, String msg, String format, String m) {
        Viewable v = checkLogin();
        if (v != null) {
            return Response.ok().entity(v).build();
        }
        int move;
        try {
            move = Integer.parseInt(m);
        } catch (NumberFormatException ex) {
            move = -1;
        }
        WebResource url = base.path("games").queryParam("loginID", uuid).path(id);
        if (move >= 0) {
            url = url.queryParam("move", "" + move);
        }
        ResponseBuilder resp = Response.ok();
        CacheControl cc = new CacheControl();
        cc.setNoCache(true);
        resp.cacheControl(cc);
        Cookie cFormat = headers.getCookies().get("format");
        if (format.length() == 0) {
            if (cFormat != null) {
                format = cFormat.getValue();
            } else {
                if (isMobile(headers)) {
                    format = "small";
                }
            }
        } else {
            if (cFormat == null || !format.equals(cFormat.getValue())) {
                resp.cookie(new NewCookie("format", format));
            }
        }

        Document doc = url.accept(MediaType.TEXT_XML).get(Document.class);
        Board b;
        String t = doc.getElementsByTagName("board").item(0).getTextContent();
        try {
            b = Board.valueOf(doc.getElementsByTagName("board").item(0).getTextContent());
        } catch (Exception ex) {
//            b = new Board(t);
//        } catch (IllegalStateException ex) {
            Exceptions.printStackTrace(ex);
            b = Board.empty();
        }
        String bCode = null;
        try{
            bCode = stat.path("openings").path(b.getCode()+".check").queryParam("loginID", user.getId()).accept(MediaType.TEXT_PLAIN).get(String.class);
        }catch(Exception e){
            bCode = null;
        }
        if(bCode == null || "".equals(bCode))
            v = viewable("game.fmt", doc, "message", msg, "format", format, "board", b,"textPicture", boardToPicture(b));
        else
            v = viewable("game.fmt", doc, "message", msg, "format", format, "board", b,"textPicture", boardToPicture(b),"bCode", bCode);
        return resp.entity(v).build();
    }

    private static String boardToPicture(Board b) {
        StringWriter w = new StringWriter();
        try {
            b.write(w);
        } catch (IOException ex) {
            return ex.toString();
        }
        return w.toString();
    }

    @GET
    @Path("games/{id}/move")
    @Produces(MediaType.TEXT_HTML)
    public Response move(
        @PathParam("id") String id,
        @QueryParam("type") String type,
        @QueryParam("direction") String direction,
        @QueryParam("direction-next") @DefaultValue("") String directionNext,
        @QueryParam("column") @DefaultValue("") String column,
        @QueryParam("row") @DefaultValue("") String row
    ) {
        Viewable v = checkLogin();
        if (v != null) {
            return Response.ok().entity(v).build();
        }
        WebResource wr = base.path("games").path(id).
            queryParam("loginID", uuid).
            queryParam("player", user.getId());
        try {
            if (type.equals("resign")) {
                wr.queryParam("move", "RESIGN").put();
                return board(id);
            }
            if (type.equals("fence")) {
                wr.queryParam("move", direction.charAt(0) + column + row).put();
                return board(id);
            }
            if (type.equals("move")) {
                wr.queryParam("move", direction + directionNext).put();
                return board(id);
            }
        } catch (UniformInterfaceException ex) {
            return board(id, "WRONG_MOVE", "-1");
        }
        return board(id);
    }

    @GET
    @Path("games/{id}/comment")
    @Produces(MediaType.TEXT_HTML)
    public Response comment(
        @PathParam("id") String id,
        @QueryParam("comment") String comment
    ) {
        Viewable v = checkLogin();
        if (v != null) {
            return Response.ok().entity(v).build();
        }
        WebResource wr = base.path("games").path(id).
            queryParam("loginID", uuid).
            queryParam("player", user.getId()).
            queryParam("comment", comment);
        wr.put();
        
        return board(id);
    }

    @GET
    @Path("games/create")
    @Produces(MediaType.TEXT_HTML)
    public Response create(
        @QueryParam("white") String white,
        @QueryParam("black") String black
    ) {
        Viewable v = checkLogin();
        if (v != null) {
            return Response.status(Response.Status.FORBIDDEN).entity(v).build();
        }

        if (user.getId().equals(white) || user.getId().equals(black)) {
            Object obj =
                base.path("games").
                queryParam("loginID", uuid).
                queryParam("white", white).
                queryParam("black", black).accept(MediaType.TEXT_XML).post(Document.class);
            return Response.ok(welcomeImpl()).build();
        } else {
            return Response.status(Response.Status.NOT_FOUND).
                entity(welcomeImpl("message", "You (" + user.getId() + ") must be white or black!")).build();
        }
    }

    private Viewable welcomeImpl(Object... args) {
        final Document got = base.path("games").queryParam("loginID", uuid).accept(MediaType.TEXT_XML).get(Document.class);
        return viewable("index.fmt", got, args);
    }

    @Path("requests")
    public Requests getRequests() {
        if (requests == null) {
            requests = new Requests(web.path("requests"));
        }
        return requests;
    }


    @GET
    @Path("options")
    public Response changeOptions(
        @QueryParam("email") String email,
        @QueryParam("language") String language,
        @QueryParam("verified") String verified
    ) throws IOException {
        Viewable v = checkLogin();
        if (v != null) {
            return Response.status(Response.Status.FORBIDDEN).entity(v).build();
        }

        if (email != null) {
            if (getRequests().isVerified(verified)) {
                UserInfo ui = base.path("users/" + user.getId()).
                    queryParam("loginID", uuid).
                    queryParam("name", "email").
                    queryParam("value", email).accept(MediaType.TEXT_XML).post(UserInfo.class);
            } else {
                WebResource request;
                request = web.path("options").queryParam("name", "email").queryParam("email", email);
                URI callback = getRequests().register(login(), request);

                ResourceBundle rb = bundle(null);
                String subject = rb.getString("MSG_ChangeEmailSubject");
                String text = MessageFormat.format(rb.getString("MSG_ChangeEmailText"), user.getId(), callback);
                EmailService.getDefault().sendEmail(email, subject, text);
                return Response.ok(viewable("email.fmt", null)).build();

            }
        }

        if (language != null) {
            UserInfo ui = base.path("users/" + user.getId()).
                queryParam("loginID", uuid).
                queryParam("name", "language").
                queryParam("value", language).
                accept(MediaType.TEXT_XML).post(UserInfo.class);
        }

        return welcome(10);
    }

    @GET
    @Path("elo")
    @Produces(MediaType.TEXT_HTML)
    public Response getEloList(
            @QueryParam("historyId") @DefaultValue("0") Integer historyId){
        Viewable v = checkLogin();
        if (v != null) {
            return Response.status(Response.Status.FORBIDDEN).entity(v).build();
        }
        final Document got = stat.path("elo").path("list").path(historyId.toString()).accept(MediaType.TEXT_XML).get(Document.class);
        return Response.ok(viewable("elo.fmt", got, "historyId", historyId)).build();
    }
    
    @GET
    @Path("openings")
    @Produces(MediaType.TEXT_HTML)
    public Response getOpeningRoot(){
        return getOpeningNode("ROOT");
    }

    @GET
    @Path("openings/{code}")
    @Produces(MediaType.TEXT_HTML)
    public Response getOpeningNode(@PathParam("code") String code){
        Viewable v = checkLogin();
        if (v != null) {
            return Response.status(Response.Status.FORBIDDEN).entity(v).build();
        }
        final Document got = stat.path("openings").path(code).queryParam("loginID", user.getId()).accept(MediaType.TEXT_XML).get(Document.class);
        Board b;
        try {
            b = Board.valueOf(got.getElementsByTagName("nodeCode").item(0).getTextContent());
        } catch (Exception ex) {
            Exceptions.printStackTrace(ex);
            b = Board.empty();
        }
        return Response.ok(viewable("openings.fmt", got, "whitefences",b.getPlayers().get(0).getFences(),"blackfences",b.getPlayers().get(1).getFences())).build();
    }

    @GET
    @Path("openings/{code}/{status}")
    @Produces(MediaType.TEXT_HTML)
    public Response getOpeningNodeGames(@PathParam("code") String code, @PathParam("status") String status){
        Viewable v = checkLogin();
        if (v != null) {
            return Response.status(Response.Status.FORBIDDEN).entity(v).build();
        }
        final Document got = stat.path("openings").path(code).path(status).queryParam("loginID", user.getId()).accept(MediaType.TEXT_XML).get(Document.class);
        return Response.ok(viewable("opening_games.fmt", got,"code",code,"color",status)).build();
    }

    @GET
    @Path("openings/{code}.png")
    @Produces("image/png")
    public Response getOpeningBoardImage(
        @PathParam("code") String code,
        @QueryParam("fieldSize") @DefaultValue("40") int fieldSize
    ) throws IllegalPositionException {
        Board b = Board.valueOf(code);
        ResponseBuilder resp = Response.ok();
        CacheControl cc = new CacheControl();
        cc.setNoCache(true);
        resp.cacheControl(cc);
        return resp.entity(BoardImage.draw(b, fieldSize)).build();
    }

    //
    // start the server
    //

    public static void main(String[] params) throws Exception {
        List<String> args = new ArrayList<String>(Arrays.asList(params));

        String publicURL = null;
        if (args.size() >= 2 && args.get(0).equals("--url")) {
            publicURL = args.get(1);
            args.remove(0);
            args.remove(0);
        }

        int port = 9333;
        if (args.size() > 1) {
            port = Integer.parseInt(args.get(0));
        }
        String remoteAPI = args.size() >= 2 ? args.get(1) : null;
        String remoteStatistics = args.size() >= 3 ? args.get(2) : null;

        Locale.setDefault(Locale.ROOT);

        Callable<Void> r = startServers(port, remoteAPI, remoteStatistics, publicURL);

        if (args.size() < 3 || !args.get(args.size() - 1).equals("--kill")) {
            System.out.println("Hit enter to stop it...");
            System.in.read();
        } else {
            synchronized (UI.class) {
                UI.class.wait();
            }
        }
        r.call();
        System.exit(0);
    }

    static Callable<Void> startServers(int port, String remoteAPI, String remoteStatistics, String publicURL) throws Exception {
        Client client = new Client();
        Client client1 = new Client();

        final HttpServer apiServer;
        if (remoteAPI == null) {
            throw new IllegalArgumentException("Provide URL to API server"); // NOI18N
        } else {
            base = client.resource(new URI(remoteAPI));
            apiServer = null;
        }

        if (remoteStatistics != null) {
            stat = client1.resource(new URI(remoteStatistics));
        } else {
            stat = client1.resource(new URI("http://localhost:9444"));
        }

        ResourceConfig rc = new PackagesResourceConfig(
            "cz.xelfi.quoridor.freemarkerdor"
        );

        final String baseUri = "http://localhost:" + port + "/";
        if (publicURL == null) {
            publicURL = baseUri;
        }
        final HttpServer server = HttpServerFactory.create(baseUri, rc);
        Client c3 = new Client();
        web = c3.resource(publicURL);
        server.start();
        System.out.println("Quoridor started at port " + port);

        return new Callable<Void>() {
            public Void call() throws Exception {
                if (apiServer != null) {
                    apiServer.stop(0);
                }
                server.stop(0);
                return null;
            }
        };
    }

    private ResourceBundle bundle(Locale[] locale) {
        ResourceBundle rb = null;
        String lng = user == null ? null : user.getProperty("language"); // NOI18N
        if (lng != null) {
            try {
                Locale l = new Locale(lng);
                rb = ResourceBundle.getBundle("cz.xelfi.quoridor.freemarkerdor.UI.Bundle", l);
                if (locale != null) {
                    locale[0] = l;
                }
            } catch (MissingResourceException e) {
                // OK
            }
        }
        if (rb == null) {
            try {
                for (Locale l : headers.getAcceptableLanguages()) {
                    try {
                        rb = ResourceBundle.getBundle("cz.xelfi.quoridor.freemarkerdor.UI.Bundle", l);
                        if (locale != null) {
                            locale[0] = l;
                        }
                        break;
                    } catch (MissingResourceException e) {
                        // OK
                    }
                }
            } catch (WebApplicationException ex) {
                // OK, can't parse header
            }
        }
        if (rb == null) {
            rb = ResourceBundle.getBundle("cz.xelfi.quoridor.freemarkerdor.UI.Bundle", Locale.ENGLISH);
            if (locale != null) {
                locale[0] = Locale.ENGLISH;
            }
        }
        return rb;
    }

    private Viewable viewable(String page, Document doc, Object... more) {
        Locale[] locale = new Locale[1];
        ResourceBundle rb = bundle(locale);

        Map<String,Object> map = new HashMap<String,Object>();
        class ConvertToDate extends HashMap<Object,Object> {
            @Override
            public Object get(Object o) {
                long time = Long.parseLong(o.toString());
                return new Date(time);
            }
        }

        map.put("locale", locale[0].toString());
        map.put("doc", doc);
        if (user != null) {
            map.put("user", user.getId());
            map.put("email", user.getProperty("email"));
        }
        map.put("bundle", rb);
        map.put("toDate", new ConvertToDate());
        map.put("now", System.currentTimeMillis());
        map.put("version", version);
        for (int i = 0; i < more.length; i += 2) {
            map.put((String)more[i],more[i + 1]);
        }
        return new Viewable(page, map);
    }


    private static boolean isMobile(HttpHeaders headers) {
        final String[] keywords = {
            "Profile/MIDP",
        };
        List<String> agent = headers.getRequestHeader(HttpHeaders.USER_AGENT);
        if (agent != null) {
            for (String a : agent) {
                for (String k : keywords) {
                    if (a.contains(k)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

}
