/*
 * Quoridor server and related libraries
 * Copyright (C) 2009-2010 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://www.gnu.org/licenses/.
 */

package cz.xelfi.quoridor.statistics;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.Collections;
import java.util.ArrayList;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
/**
 *
 * @author Martin Rexa
 */
@XmlRootElement
@XmlAccessorType(XmlAccessType.FIELD)
public class EloList {
    static int INITIAL_ELO = 1250;
    static int MAX_ELO_DIFF = 400;
    static int KVAL_START = 25;
    static int KVAL_FOLLOW = 10;
    static int KVAL_LIMIT = 30;

    Map<String, Double> players;
    Map<String, Integer> playerGames;
    long published = 0;
    boolean originalList;


    public EloList () {
        players = new HashMap<String, Double>();
        playerGames = new HashMap<String, Integer>();
        originalList = true;
    }

    public EloList(EloList src){
        this.players = new HashMap<String, Double>(src.players);
        this.playerGames = new HashMap<String, Integer>(src.playerGames);
        originalList = false;
    }

    public void setEndDate(long endDate){
        this.published = endDate;
    }

    public long getEndDate(){
        return this.published;
    }

    @XmlElementWrapper(name="elolist")
    @XmlElement(name="item")
    public List<EloEntry> getFinalList(){
        List<EloEntry> finalList = new ArrayList<EloEntry>();
        for (Map.Entry<String, Double> player : players.entrySet()){
            finalList.add(new EloEntry(player.getKey(), player.getValue(), playerGames.get(player.getKey())));
        }
        Collections.sort(finalList, EloEntry.BEST_FIRST);
        return finalList;
    }

    public EloList putResult(String winner, String looser){
        double wElo = getElo(winner);
        double lElo = getElo(looser);

        double eloDiff = wElo - lElo;
        if(eloDiff < - MAX_ELO_DIFF)
            eloDiff = - MAX_ELO_DIFF;
        if(eloDiff > MAX_ELO_DIFF)
            eloDiff = MAX_ELO_DIFF;

        double wDiff = Math.round(100 * getKVal(winner) * (1 - (1 / (1 + Math.pow(10, -eloDiff / MAX_ELO_DIFF)))));
        wDiff /= 100.0;
        double lDiff = Math.round(100 * getKVal(looser) * (0 - (1 / (1 + Math.pow(10, eloDiff / MAX_ELO_DIFF)))));
        lDiff /= 100.0;

        players.put(winner, wElo + wDiff);
        players.put(looser, lElo + lDiff);
        return this;
    }
    
    public double getElo(String player){
        Double elo = players.get(player);
        return (elo == null) ? INITIAL_ELO : elo;
    }
    
    private int getKVal(String player){
        Integer games = playerGames.get(player);
        int g = (games == null) ? 0 : games;
        // Side effect, method should not be called more times for the same game/player
        playerGames.put(player, g + 1);
        return g < KVAL_LIMIT ? KVAL_START : KVAL_FOLLOW;
    }

}
